LX_POS_TAGS = {
    "ADJ",
    "ADV",
    "CARD",
    "CJ",
    "CL",
    "CN",
    "DA",
    "DEM",
    "DFR",
    "DGT",
    "DGTR",
    "DM",
    "EADR",
    "EL",
    "EMP",
    "EOE",
    "EXC",
    "GER",
    "IA",
    "IND",
    "INF",
    "INT",
    "ITJ",
    "LADV1",
    "LADV2",
    "LADV3",
    "LADV4",
    "LADV5",
    "LADV6",
    "LADV7",
    "LCJ1",
    "LCJ2",
    "LCJ3",
    "LCJ4",
    "LCJ5",
    "LD1",
    "LD2",
    "LDEM1",
    "LDEM2",
    "LDFR1",
    "LDFR2",
    "LDM1",
    "LDM2",
    "LDM3",
    "LINT1",
    "LINT2",
    "LITJ1",
    "LITJ2",
    "LITJ3",
    "LPREP1",
    "LPREP2",
    "LPREP3",
    "LPREP4",
    "LPREP5",
    "LPREP6",
    "LPRS1",
    "LPRS2",
    "LQD1",
    "LQD2",
    "LREL1",
    "LREL2",
    "LTR",
    "MGT",
    "MTH",
    "NP",
    "NULL",
    "ORD",
    "PADR",
    "PL",
    "PNM",
    "PNT",
    "POSS",
    "PP",
    "PPA",
    "PPT",
    "PREP",
    "PRS",
    "QNT",
    "REL",
    "SC",
    "STT",
    "SYB",
    "TERMN",
    "UM",
    "UNIT",
    "V",
    "VAUX",
    "VAUXGER",
    "VAUXINF",
    "WD",
    "X",
}

U_POS_TAGS = {
    "ADJ",
    "ADP",
    "ADV",
    "AUX",
    "CCONJ",
    "DET",
    "INT",
    "INTJ",
    "NOUN",
    "NUM",
    "PRON",
    "PROPN",
    "PUNCT",
    "SCONJ",
    "SYM",
    "VERB",
    "X",
}

LX_DEPREL_TAGS = {
    "ADV",
    "C-ARG1",
    "C-ARG2",
    "C-ARG3",
    "C",
    "CARD",
    "CONJ",
    "COORD",
    "DEM",
    "DEP",
    "DM",
    "DO-ARG1",
    "DO-ARG2",
    "DO-ARG21",
    "DO-ARG3",
    "DO-TMP",
    "DO",
    "ERROR",
    "INT",
    "IO-ARG2",
    "IO-ARG3",
    "IO-ARG31",
    "ITJ",
    "M-ADV",
    "M-ARG1",
    "M-ARG2",
    "M-CAU",
    "M-DIR",
    "M-EXT",
    "M-LOC",
    "M-M",
    "M-MNR",
    "M-PNC",
    "M-POV",
    "M-PRED",
    "M-TMP",
    "M",
    "N",
    "OBL-ADV",
    "OBL-ARG1",
    "OBL-ARG2",
    "OBL-ARG3",
    "OBL-ARG4",
    "ORD",
    "P",
    "PERCENT",
    "POSS",
    "PRD-ARG2",
    "PRD-ARG3",
    "PRD-ARG4",
    "PRD",
    "PRS",
    "PUNCT",
    "QNT",
    "REL",
    "ROOT",
    "S",
    "SJ-ARG1",
    "SJ-ARG11",
    "SJ-ARG11cp",
    "SJ-ARG1cp",
    "SJ-ARG2",
    "SJ-ARG21cp",
    "SJ-ARG2ac",
    "SJ-ARG2cp",
    "SJ",
    "SP-ARG1",
    "SP-ARG2",
    "SP-M",
    "SP-PRED",
    "SP",
    "V",
}

U_DEPREL_TAGS = {
    # only contains UD V2 tags that occur in CINTIL, i.e. tags known to our DepParser
    "advcl",
    "advmod",
    "amod",
    "appos",
    "aux",
    "case",
    "cc",
    "ccomp",
    "comp",
    "conj",
    "cop",
    "csubj:pass",
    "csubj",
    "dep",
    "det",
    "fixed",
    "iobj",
    "mark",
    "mod",
    "nsubj",
    "num",
    "number",
    "obj",
    "parataxis",
    "pobj",
    "poss",
    "preconj",
    "predet",
    "punct",
    "root",
    "xcomp",
}

U_DEPREL_TAGS_V1_TO_V2 = {
    # https://universaldependencies.org/v2/summary.html#clauses-and-dependents-of-predicates
    "dobj": "obj",
    "nsubjpass": "nsubj:pass",  # does not occur in CINTIL
    "csubjpass": "csubj:pass",
    "auxpass": "aux:pass",  # does not occur in CINTIL
    # https://universaldependencies.org/v2/summary.html#multiword-expressions
    "mwe": "fixed",
}

LX_FEATS = {
    "1p",
    "1s",
    "2p",
    "2s",
    "3p",
    "3s",
    "c-1p",
    "c-1s",
    "c-2p",
    "c-2s",
    "c-3p",
    "c-3s",
    "fc-1p",
    "fc-1s",
    "fc-2s",
    "fc-3p",
    "fc-3s",
    "fi-1p",
    "fi-1s",
    "fi-2p",
    "fi-2s",
    "fi-3p",
    "fi-3s",
    "fp-comp",
    "fp-dim",
    "fp-sup",
    "fp",
    "fp1",
    "fp2",
    "fp3",
    "fs-comp-dim",
    "fs-comp",
    "fs-dim",
    "fs-sup",
    "fs",
    "fs1",
    "fs2",
    "fs3",
    "gn",
    "gn13",
    "gn3",
    "gp-comp",
    "gp",
    "gp1",
    "gp2",
    "gp3",
    "gs-comp",
    "gs-sup",
    "gs",
    "gs1",
    "gs2",
    "gs3",
    "ic-1p",
    "ic-1s",
    "ic-2s",
    "ic-3p",
    "ic-3s",
    "ii-1p",
    "ii-1s",
    "ii-2s",
    "ii-3p",
    "ii-3s",
    "imp-1p",
    "imp-2p",
    "imp-2s",
    "imp-Cp",
    "imp-Cs",
    "impaf-2p",
    "impaf-2s",
    "impneg-2p",
    "impneg-2s",
    "mn",
    "mp-comp",
    "mp-dim",
    "mp-sup",
    "mp",
    "mp1",
    "mp2",
    "mp3",
    "mpC",
    "mpi-1p",
    "mpi-1s",
    "mpi-3p",
    "mpi-3s",
    "ms-comp-dim",
    "ms-comp",
    "ms-dim",
    "ms-sup-dim",
    "ms-sup",
    "ms",
    "ms1",
    "ms2",
    "ms3",
    "ndef",
    "ninf",
    "np",
    "pc-1p",
    "pc-1s",
    "pc-2p",
    "pc-2s",
    "pc-3p",
    "pc-3s",
    "pi-1p",
    "pi-1s",
    "pi-2p",
    "pi-2s",
    "pi-3p",
    "pi-3s",
    "pi-Cp",
    "ppi-1p",
    "ppi-1s",
    "ppi-2p",
    "ppi-2s",
    "ppi-3p",
    "ppi-3s",
}

# This was obtained from values of lx2u.feats.LX_FEATS_TO_U_FEATS.
# The ones that pass through _get_add_verb_form_inf_callable appear twice, once
# as they are, another with |VerbForm=Inf appended to them.
# The resulting list has been sorted and de-duplicated.
U_FEATS = {
    "Degree=Abs|Gender=Fem|Number=Plur",
    "Degree=Abs|Gender=Fem|Number=Sing",
    "Degree=Cmp|Gender=Fem|Number=Plur",
    "Degree=Cmp|Gender=Fem|Number=Sing",
    "Gender=Fem|Number=Plur",
    "Gender=Fem|Number=Plur|Person=1",
    "Gender=Fem|Number=Plur|Person=2",
    "Gender=Fem|Number=Plur|Person=3",
    "Gender=Fem|Number=Sing",
    "Gender=Fem|Number=Sing|Person=1",
    "Gender=Fem|Number=Sing|Person=2",
    "Gender=Fem|Number=Sing|Person=3",
    "Gender=Masc",
    "Gender=Masc|Number=Plur",
    "Gender=Masc|Number=Plur|Degree=Abs",
    "Gender=Masc|Number=Plur|Degree=Cmp",
    "Gender=Masc|Number=Plur|Person=1",
    "Gender=Masc|Number=Plur|Person=2",
    "Gender=Masc|Number=Plur|Person=2|Polite=Form",
    "Gender=Masc|Number=Plur|Person=3",
    "Gender=Masc|Number=Sing",
    "Gender=Masc|Number=Sing|Degree=Abs",
    "Gender=Masc|Number=Sing|Degree=Cmp",
    "Gender=Masc|Number=Sing|Person=1",
    "Gender=Masc|Number=Sing|Person=2",
    "Gender=Masc|Number=Sing|Person=3",
    "Mood=Cnd|Number=Plur|Person=1",
    "Mood=Cnd|Number=Plur|Person=2",
    "Mood=Cnd|Number=Plur|Person=3",
    "Mood=Cnd|Number=Sing|Person=1",
    "Mood=Cnd|Number=Sing|Person=2",
    "Mood=Cnd|Number=Sing|Person=3",
    "Mood=Imp|Person=1|Number=Plur",
    "Mood=Imp|Person=2|Number=Plur",
    "Mood=Imp|Person=2|Number=Sing",
    "Mood=Imp|Person=2|Polite=Form|Number=Plur",
    "Mood=Imp|Person=2|Polite=Form|Number=Sing",
    "Mood=Ind|Number=Plur|Person=1|Tense=Fut",
    "Mood=Ind|Number=Plur|Person=1|Tense=Past",
    "Mood=Ind|Number=Plur|Person=1|Tense=Pres",
    "Mood=Ind|Number=Plur|Person=2|Polite=Form|Tense=Pres",
    "Mood=Ind|Number=Plur|Person=2|Tense=Fut",
    "Mood=Ind|Number=Plur|Person=2|Tense=Past",
    "Mood=Ind|Number=Plur|Person=2|Tense=Pres",
    "Mood=Ind|Number=Plur|Person=3|Tense=Fut",
    "Mood=Ind|Number=Plur|Person=3|Tense=Past",
    "Mood=Ind|Number=Plur|Person=3|Tense=Pres",
    "Mood=Ind|Number=Sing|Person=1|Tense=Fut",
    "Mood=Ind|Number=Sing|Person=1|Tense=Past",
    "Mood=Ind|Number=Sing|Person=1|Tense=Pres",
    "Mood=Ind|Number=Sing|Person=2|Tense=Fut",
    "Mood=Ind|Number=Sing|Person=2|Tense=Past",
    "Mood=Ind|Number=Sing|Person=2|Tense=Pres",
    "Mood=Ind|Number=Sing|Person=3|Tense=Fut",
    "Mood=Ind|Number=Sing|Person=3|Tense=Past",
    "Mood=Ind|Number=Sing|Person=3|Tense=Pres",
    "Mood=Sub|Number=Plur|Person=1|Tense=Fut",
    "Mood=Sub|Number=Plur|Person=1|Tense=Pres",
    "Mood=Sub|Number=Plur|Person=2|Tense=Pres",
    "Mood=Sub|Number=Plur|Person=3|Tense=Fut",
    "Mood=Sub|Number=Plur|Person=3|Tense=Pres",
    "Mood=Sub|Number=Sing|Person=1|Tense=Fut",
    "Mood=Sub|Number=Sing|Person=1|Tense=Pres",
    "Mood=Sub|Number=Sing|Person=2|Tense=Fut",
    "Mood=Sub|Number=Sing|Person=2|Tense=Pres",
    "Mood=Sub|Number=Sing|Person=3|Tense=Fut",
    "Mood=Sub|Number=Sing|Person=3|Tense=Pres",
    "Number=Plur",
    "Number=Plur|Degree=Cmp",
    "Number=Plur|Person=1",
    "Number=Plur|Person=1|VerbForm=Inf",
    "Number=Plur|Person=2",
    "Number=Plur|Person=2|VerbForm=Inf",
    "Number=Plur|Person=3",
    "Number=Plur|Person=3|VerbForm=Inf",
    "Number=Sing",
    "Number=Sing|Degree=Abs",
    "Number=Sing|Degree=Cmp",
    "Number=Sing|Person=1",
    "Number=Sing|Person=1|VerbForm=Inf",
    "Number=Sing|Person=2",
    "Number=Sing|Person=2|VerbForm=Inf",
    "Number=Sing|Person=3",
    "Number=Sing|Person=3|VerbForm=Inf",
    "Person=1",
    "Person=2",
    "Person=3",
    "Tense=Imp|Mood=Ind|Person=1|Number=Plur",
    "Tense=Imp|Mood=Ind|Person=1|Number=Sing",
    "Tense=Imp|Mood=Ind|Person=2|Number=Sing",
    "Tense=Imp|Mood=Ind|Person=3|Number=Plur",
    "Tense=Imp|Mood=Ind|Person=3|Number=Sing",
    "Tense=Imp|Mood=Sub|Person=1|Number=Plur",
    "Tense=Imp|Mood=Sub|Person=1|Number=Sing",
    "Tense=Imp|Mood=Sub|Person=2|Number=Sing",
    "Tense=Imp|Mood=Sub|Person=3|Number=Plur",
    "Tense=Imp|Mood=Sub|Person=3|Number=Sing",
    "Tense=Pqp|Mood=Ind|Person=1|Number=Plur",
    "Tense=Pqp|Mood=Ind|Person=1|Number=Sing",
    "Tense=Pqp|Mood=Ind|Person=3|Number=Plur",
    "Tense=Pqp|Mood=Ind|Person=3|Number=Sing",
    "VerbForm=Inf",
}

NE_TAGS = {
    "B-EVT",
    "B-LOC",
    "B-MSC",
    "B-ORG",
    "B-PER",
    "B-WRK",
    "I-EVT",
    "I-LOC",
    "I-MSC",
    "I-ORG",
    "I-PER",
    "I-WRK",
    "O",
}



"""
Luís Gomes <luis.gomes@di.fc.ul.pt>, <luismsgomes@gmail.com>

Copyright NLX-Group, Universidade de Lisboa, 2015-2020
"""

__version__ = "2.13.0"


from .cintilformat import (  # noqa: F401
    CintilFormatError,
    CintilFormatSpec,
    CintilTokenMissingPOS,
    CintilTokenMissingInfl,
    CintilTokenMissingNE,
    CintilSentenceMissingTag,
    CintilSentenceExtraneousTag,
    CintilParagraphMissingTag,
    CintilParagraphExtraneousTag,
    CintilParagraphExtraneousText,
    format_cintil_paragraph,
    format_cintil_sentence,
    format_cintil_token,
    parse_cintil_paragraph,
    parse_cintil_sentence,
    parse_cintil_token,
    read_cintil_paragraphs,
    write_cintil_paragraphs,
)
from .conllformat import format_conll_sentence  # noqa: F401
from .jsonlformat import (  # noqa: F401
    jsonsentence_to_lxsentence,
    lxsentence_to_jsonsentence,
    read_jsonl_paragraphs,
    read_jsonl_sentences,
    write_jsonl_paragraphs,
    write_jsonl_paragraph,
    write_jsonl_sentence,
)
from .lxtoken import LxToken  # noqa: F401
from .lxsentence import LxSentence  # noqa: F401
from .lxparagraph import LxParagraph  # noqa: F401
from .contractions import CONTRACTIONS  # noqa: F401
from .utils import normalize_unicode, normalize_text, recase_as  # noqa: F401



ESCAPE_BRACKETS = {
    "(": "-lrb-",
    ")": "-rrb-",
    "[": "-lsb-",
    "]": "-rsb-",
    "{": "-lcb-",
    "}": "-rcb-",
}

UNESCAPE_BRACKETS = {
    "-lrb-": "(",
    "-rrb-": ")",
    "-lsb-": "[",
    "-rsb-": "]",
    "-lcb-": "{",
    "-rcb-": "}",
}


def escape_token(token):
    return ESCAPE_BRACKETS.get(token, token)


def unescape_token(token):
    return UNESCAPE_BRACKETS.get(token, token)



"Module for reading legacy format LXTSV (similar to CONLL)"

from lxcommon.lxsentence import LxSentence


def read_lxtsv(lines):
    """Reads and yields sentences as Python lists of dicts (each token is a dict)"""
    enumerated_lines = enumerate(lines, start=1)

    expected_column_names = [
        "form",
        "raw",
        "lemma",
        "pos",
        "infl",
        "ne",
        "deprel",
        "parent",
        "udeprel",
        "uparent",
        "space",
    ]

    # first line should contain column names; sanity check here:
    for linenum, line in enumerated_lines:
        column_names = line.rstrip("\n").split("\t")
        if column_names != expected_column_names:
            raise Exception(
                f"Unexpected column names in first line; got {column_names}; "
                f"expected {expected_column_names}."
            )
        break
    else:  # file is empty
        return

    sentence = []
    for linenum, line in enumerated_lines:
        line = line.rstrip("\n")
        if not line:
            if sentence:
                yield sentence
                sentence = []
        else:
            columns = line.split("\t")
            if len(columns) != len(column_names):
                raise Exception(
                    f"Unexpected number of columns in line {linenum}; "
                    f"expected {len(column_names)}, got {len(columns)}."
                )
            sentence.append(dict(zip(column_names, columns)))
    if sentence:
        yield sentence


def read_lxsentences_from_lxtsv(lines):
    for sentence in read_lxtsv(lines):
        yield LxSentence.from_primitive_types(sentence)



import logging
import re
import unicodedata

from .utils import recase_as, normalize_text


LOG = logging.getLogger(__name__)


class CintilFormatError(Exception):
    pass


class CintilTokenMissingPOS(CintilFormatError):
    pass


class CintilTokenMissingInfl(CintilFormatError):
    pass


class CintilTokenMissingNE(CintilFormatError):
    pass


class CintilSentenceMissingTag(CintilFormatError):
    pass


class CintilSentenceExtraneousTag(CintilFormatError):
    pass


class CintilParagraphMissingTag(CintilFormatError):
    pass


class CintilParagraphExtraneousTag(CintilFormatError):
    pass


class CintilParagraphExtraneousText(CintilFormatError):
    pass


class CintilExtraneousText(CintilFormatError):
    pass


class CintilFormatSpecError(Exception):
    pass


class CintilFormatSpec:
    def __init__(
        self,
        tokenized=None,
        pos=None,
        morph=None,
        ne=None,
        bracketed=None,
        strict=False,
    ):
        """One or all of tokenized, pos, morph or ne must be given and of type bool.

        If only tokenized is given, it implies pos=False, morph=False and ne=False
        If only pos is given, it implies tokenized=True, morph=False and ne=False
        If only morph is given, it implies tokenized=True, pos=True and ne=False
        If only ne is given, it implies tokenized=True, pos=True and morph=True

        Argument bracketed may be given only when ne=True and defaults to True.

        If strict is given and True, then exceptions will be raised whenever an
        error is found instead of returning the best effort.
        """
        arg_names = ["tokenized", "pos", "morph", "ne"]
        arg_values = (tokenized, pos, morph, ne)
        n_given = 0
        for name, value in zip(arg_names, arg_values):
            if value is not None:
                if not isinstance(value, bool):
                    raise TypeError(f"{name} must be a bool")
                n_given += 1
        if n_given == 1:
            tokenized, pos, morph, ne = {
                # given args: result
                (False, None, None, None): (False, False, False, False),
                (True, None, None, None): (True, False, False, False),
                (None, False, None, None): (True, False, False, False),
                (None, True, None, None): (True, True, False, False),
                (None, None, False, None): (True, True, False, False),
                (None, None, True, None): (True, True, True, False),
                (None, None, None, False): (True, True, True, False),
                (None, None, None, True): (True, True, True, True),
            }[arg_values]
        elif n_given == len(arg_names):
            # all were given; let's check their consistency
            # (by checking that arg_values are sorted)
            for lname, lvalue, rname, rvalue in zip(
                arg_names[:-1], arg_values[:-1], arg_names[1:], arg_values[1:]
            ):
                if lvalue > rvalue:
                    raise CintilFormatSpecError(
                        f"invalid format spec: {rname} cannot be True"
                        f" if {lname} is False"
                    )
        else:  # either none or more than one but not all arguments were given
            raise CintilFormatSpecError(
                "either one or all of tokenized, pos, morph or ne"
                " must be given and of type bool"
            )
        if ne is True:
            if bracketed is not None and not isinstance(bracketed, bool):
                raise TypeError("bracketed must be of type bool")
            bracketed = bracketed or bracketed is None
        elif bracketed is not None:
            raise CintilFormatSpecError("bracketed may be given only if ne is True")
        super(CintilFormatSpec, self).__setattr__("tokenized", tokenized)
        super(CintilFormatSpec, self).__setattr__("pos", pos)
        super(CintilFormatSpec, self).__setattr__("morph", morph)
        super(CintilFormatSpec, self).__setattr__("ne", ne)
        super(CintilFormatSpec, self).__setattr__("bracketed", bracketed)
        super(CintilFormatSpec, self).__setattr__("strict", strict)

    def __setattr__(self, name, value):
        raise CintilFormatSpecError("CintilFormatSpec objects are immutable")

    def __repr__(self):
        return (
            f"CintilFormatSpec(tokenized={self.tokenized}, pos={self.pos}, "
            f"morph={self.morph}, ne={self.ne}, bracketed={self.bracketed})"
        )

    def __eq__(self, other):
        if not isinstance(other, CintilFormatSpec):
            raise TypeError("other must be an instance of CintilFormatSpec")
        return (
            self.tokenized == other.tokenized
            and self.pos == other.pos
            and self.morph == other.morph
            and self.ne == other.ne
            and self.bracketed == other.bracketed
        )


def _maybeatag(form):
    # this may give false positives but it's better to be on the safe side
    return form.startswith("<") and form.endswith(">")


_remainder_pat = "(?P<remainder>.*)"


_ne_pat = r"(?P<ne>O|[BI]-[A-Z]+)"
_match_ne = re.compile(f"^{_remainder_pat}/{_ne_pat}$").match
_match_bracketed_ne = re.compile(f"^{_remainder_pat}\\[{_ne_pat}\\]$").match


def _parse_bracketed_ne(token):
    m = _match_bracketed_ne(token)
    if m:
        return m.group("remainder"), m.group("ne")
    return token, None


def _parse_ne(token):
    m = _match_ne(token)
    if m:
        return m.group("remainder"), m.group("ne")
    return token, None


# TODO: write a regex that matches exactly the set of possible inflection flags
# A question mark (?) is inserted by vlemma when the infl is not known
# apparently, two question marks are also allowed as in Monsenhor/STT#??
_infl_pat = r"(?P<infl>[?a-zA-Z1-9]+(?:-[?a-zA-Z1-9]+)*)"
_match_infl = re.compile(f"^{_remainder_pat}#{_infl_pat}$").match


def _parse_infl(token):
    m = _match_infl(token)
    if m:
        return m.group("remainder"), m.group("infl")
    if token.endswith("#"):
        # smelly case: inflection tags separator present but no tags given
        # let it pass as if no separator and no tags were present;
        token = token[:-1]
    return token, None


# TODO: write a regex that matches exactly the set of possible POS tags
_pos_pat = r"(?P<pos>@?[A-Z]+|L[A-Z]+[1-9]+)"
_match_pos = re.compile(f"^{_remainder_pat}/{_pos_pat}$").match


def _parse_pos(token):
    m = _match_pos(token)
    if m:
        return m.group("remainder"), m.group("pos")
    return token, None


# A lemma is either
#   - question mark (?) (sometimes inserted by vlemma when the lemma is not known)
#   - a sequence of any characters except slash ([^/]+)
_single_lemma_pat = r"(?:\?|[^/]+|@URL|@EMAIL)"
_lemma_pat = r"(?P<lemma>" + _single_lemma_pat + "(?:," + _single_lemma_pat + ")*)"
_match_lemma = re.compile(f"^{_remainder_pat}/{_lemma_pat}$").match


def _parse_lemma(token):
    m = _match_lemma(token)
    if m:
        return m.group("remainder"), m.group("lemma")
    return token, None


def _is_punct(form):
    return all(unicodedata.category(c)[0] == "P" for c in form)


def _parse_form(form):
    if form.startswith("\\*"):
        form = form[2:]
        space = "L"
    else:
        space = None
    if form.endswith("*/"):
        space = "LR" if space == "L" else "R"
        form = form[:-2]
    if space is None:
        if _is_punct(form):
            # puctuation has no spaces unless we say otherwise (space not None)
            space = ""
        else:
            # every token that is neither punctuation nor symbol has spaces on both
            # sides unless they are prefixed with a hyphen - or suffixed with an
            # underscore_
            space = "L" if not form.startswith("-") else ""
            space += "R" if not form.endswith("_") else ""
    return form, space


def parse_cintil_token(token, format_spec):
    """
    Parses a token in CINTIL format into a dictionary.
    Notes:
        format_spec.pos=False implies
            format_spec.morph=False and
            format_spec.ne=False
        format_spec.morph=False implies
            format_spec.ne=False

    Admissible formats:

        word
            format_spec.pos=False, format_spec.morph=False, format_spec.ne=False

        word/pos
            format_spec.pos=True, format_spec.morph=False, format_spec.ne=False

        word/lemma/pos#infl
            format_spec.pos=True, format_spec.morph=True, format_spec.ne=False

        word/pos#infl   (function words)
            format_spec.pos=True, format_spec.morph=True, format_spec.ne=False

        word/lemma/pos#infl/ne
            format_spec.pos=True, format_spec.morph=True, format_spec.ne=True

        word/lemma/pos#infl[ne]
            format_spec.pos=True, format_spec.morph=True, format_spec.ne=True,
            format_spec.bracketed=True

    """
    if not isinstance(token, str):
        raise TypeError("token must be a string")
    if not isinstance(format_spec, CintilFormatSpec):
        raise TypeError("format_spec must be an instance of CintilFormatSpec")

    token = normalize_text(token)

    if _maybeatag(token):
        return dict(form=token)

    d = dict()
    remainder = token

    if format_spec.ne:
        if format_spec.bracketed:
            remainder, ne = _parse_bracketed_ne(remainder)
        else:
            remainder, ne = _parse_ne(remainder)
        if not ne:
            LOG.warning(f"missing NE tag in token {token!r}")
            if format_spec.strict:
                raise CintilTokenMissingNE(f"missing NE tag in token {token!r}")
        d["ne"] = ne

    if format_spec.morph:
        remainder, infl = _parse_infl(remainder)
        if infl:
            d["infl"] = infl
        remainder, pos = _parse_pos(remainder)
        if not pos:
            LOG.warning(f"missing pos tag in token {token!r}")
            if format_spec.strict:
                raise CintilTokenMissingPOS(f"missing POS tag in token {token!r}")
        d["pos"] = pos
        remainder, lemma = _parse_lemma(remainder)
        if lemma:
            d["lemma"] = lemma
            if infl is None and pos != "EADR":
                LOG.warning(f"missing inflection tags in token {token!r}")
                if format_spec.strict:
                    raise CintilTokenMissingInfl(
                        f"missing inflection tags in token {token!r}"
                    )
            # We should always have inflection flags if we have a lemma, except
            #  if pos is EADR.
            # However, it is perfectly normal to have inflection tags without lemma.
            # for example: o/DA#ms

    elif format_spec.pos:
        remainder, pos = _parse_pos(remainder)
        if not pos:
            LOG.warning(f"missing POS tag in token {token!r}")
            if format_spec.strict:
                raise CintilTokenMissingPOS(f"missing POS tag in token {token!r}")
        d["pos"] = pos

    form, space = _parse_form(remainder)
    d["form"] = form
    d["space"] = space

    return d


def parse_cintil_sentence(sentence, format_spec):
    """Parses a sentence in CINTIL format.

    Checks that the sentence is wrapped with <s> and </s> tags and raises
    CintilFormatError if it is not.
    If format_spec.tokenized is False this function will return the sentence
    as a single string without <s> and </s> tags.
    If format_spec.tokenized is True this function will split the sentence on
    whitespace, invoke lxcommon.cintilformat.parse_token for each token, and
    return a list of dictionaries.

    """
    if not isinstance(sentence, str):
        raise TypeError("sentence must be a string")
    if not isinstance(format_spec, CintilFormatSpec):
        raise TypeError("format_spec must be an instance of CintilFormatSpec")
    original_sentence, sentence = sentence, sentence.strip()
    sentence = sentence.replace("<S>", "<s>").replace("</S>", "</s>")
    if not sentence.startswith("<s>") or not sentence.endswith("</s>"):
        msg = f"sentence is not wrapped within <s> and </s> tags: {original_sentence!r}"
        LOG.warning(msg)
        if format_spec.strict:
            raise CintilSentenceMissingTag(msg)
        # else just assume that this is a sentence (best effort)
    if sentence.startswith("<s>"):
        sentence = sentence[3:].lstrip()
    if sentence.endswith("</s>"):
        sentence = sentence[:-4].rstrip()
    if "<s>" in sentence or "</s>" in sentence:
        msg = f"sentence contains multiple <s> or </s> tags: {original_sentence!r}"
        LOG.warning(msg)
        if format_spec.strict:
            raise CintilSentenceExtraneousTag(msg)
        else:
            # delete extraneous tags (best effort)
            sentence = sentence.replace("<s>", "").replace("</s>", "")
    if format_spec.tokenized:
        sentence = [
            parse_cintil_token(token, format_spec) for token in sentence.split()
        ]
        for previous_token, token in zip(sentence, sentence[1:]):
            if "R" not in previous_token["space"] and "L" in token["space"]:
                token["space"] = token["space"].replace("L", "")
            elif "L" not in token["space"] and "R" in previous_token["space"]:
                previous_token["space"] = previous_token["space"].replace("R", "")
    return sentence


def parse_cintil_paragraph(paragraph, format_spec):
    """Parses a paragraph in CINTIL format.

    Checks that the paragraph is wrapped with <p> and </p> tags and raises
    CintilFormatError if it is not.
    If format_spec.tokenized is False this function will return the paragraph
    as a list of strings, one for each sentence in the paragraph.
    If format_spec.tokenized is True this function will return a list of lists
    of dictionaries.  Each dictionary represents a token and each list of
    dictionaries represents a sentence.

    """
    if not isinstance(paragraph, str):
        raise TypeError("paragraph must be a string")
    if not isinstance(format_spec, CintilFormatSpec):
        raise TypeError("format_spec must be an instance of CintilFormatSpec")
    original_paragraph, paragraph = paragraph, paragraph.strip()
    paragraph = (
        paragraph.replace("<P>", "<p>")
        .replace("</P>", "</p>")
        .replace("<S>", "<s>")
        .replace("</S>", "</s>")
    )
    if not paragraph.startswith("<p>") or not paragraph.endswith("</p>"):
        msg = (
            f"paragraph is not wrapped within <p> and </p> tags: {original_paragraph!r}"
        )
        LOG.warning(msg)
        if format_spec.strict:
            raise CintilParagraphMissingTag(msg)
        # else just assume that this is a paragraph (best effort)
    if paragraph.startswith("<p>"):
        paragraph = paragraph[3:].lstrip()
    if paragraph.endswith("</p>"):
        paragraph = paragraph[:-4].rstrip()
    if "<p>" in paragraph or "</p>" in paragraph:
        msg = f"paragraph contains multiple <p> or </p> tags: {original_paragraph!r}"
        LOG.warning(msg)
        if format_spec.strict:
            raise CintilParagraphExtraneousTag(msg)
        else:
            # delete extraneous tags (best effort)
            paragraph = paragraph.replace("<p>", "").replace("</p>", "")

    sentences = []
    warning_shown = False
    while paragraph:
        end = paragraph.find("</s>")
        if not paragraph.startswith("<s>") or end == -1:
            msg = (
                "paragraph contains text outside <s> and </s> tags:"
                f" {original_paragraph!r}"
            )
            if not warning_shown:
                LOG.warning(msg)
                warning_shown = True
            if format_spec.strict:
                raise CintilParagraphExtraneousText(msg)
            if end == -1:  # no </s>, let's try to split on next <s> (best effort)
                if paragraph.startswith("<s>"):
                    # remove "<s>" from begining of paragraph:
                    paragraph = paragraph[3:]
                # consider the next "<s>" as the marker of end of this sentence:
                end = paragraph.find("<s>")
                if end == -1:
                    sentence, paragraph = paragraph, ""
                else:
                    sentence, paragraph = paragraph[:end].strip(), paragraph[end:]
                continue
            # else this means paragraph does not start with <s> but end != -1
        sentence, paragraph = paragraph[: end + 4], paragraph[end + 4 :].lstrip()
        sentences.append(parse_cintil_sentence(sentence, format_spec))
    return sentences


def format_cintil_token(token, format_spec):
    "Formats a token in CINTIL format."
    if not isinstance(token, dict):
        raise TypeError("token must be a dict")
    if not isinstance(format_spec, CintilFormatSpec):
        raise TypeError("format_spec must be an instance of CintilFormatSpec")

    form = recase_as(token["form"], token["raw"]) if "raw" in token else token["form"]
    form = normalize_text(form)
    if _maybeatag(form):
        return form

    cintil_token = form

    space = token.get("space")
    if _is_punct(form) and space:
        if "L" in space:
            cintil_token = r"\*" + cintil_token
        if "R" in space:
            cintil_token += "*/"

    if not format_spec.pos:
        return cintil_token

    lemma = token.get("lemma") if format_spec.morph else None
    if lemma:
        lemma = normalize_text(lemma)
        cintil_token += f"/{lemma}"

    pos = token.get("pos")
    cintil_token += f"/{pos}"

    infl = None
    if format_spec.morph:
        infl = token.get("feats", token.get("infl", None))
    if infl:
        cintil_token += f"#{infl}"

    if format_spec.ne:
        ne = token.get("ne") or token.get("ne_rb")
        cintil_token += f"[{ne}]" if format_spec.bracketed else f"/{ne}"

    return cintil_token


def format_cintil_sentence(sentence, format_spec):
    if not isinstance(format_spec, CintilFormatSpec):
        raise TypeError("format_spec must be an instance of CintilFormatSpec")
    if not format_spec.tokenized:
        if not isinstance(sentence, str):
            raise TypeError(
                "sentence must be a string (because format_spec.tokenized is False)"
            )
        sentence = sentence.strip()
        return f"<s> {sentence} </s>" if sentence else "<s> </s>"
    if not isinstance(sentence, (list, tuple)):
        raise TypeError("sentence must be an instance of list or tuple")
    if not sentence:
        return "<s> </s>"
    cintil_tokens = [format_cintil_token(token, format_spec) for token in sentence]
    return "<s> " + " ".join(cintil_tokens) + " </s>"


def format_cintil_paragraph(paragraph, format_spec):
    if not isinstance(paragraph, (list, tuple)):
        raise TypeError("paragraph must be an instance of list or tuple")
    if not isinstance(format_spec, CintilFormatSpec):
        raise TypeError("format_spec must be an instance of CintilFormatSpec")
    if not paragraph:
        return "<p> </p>"
    cintil_sentences = [
        format_cintil_sentence(sentence, format_spec) for sentence in paragraph
    ]
    return "<p> " + "\n".join(cintil_sentences) + " </p>"


def _get_close_p_pos(s):
    pos = s.find("</p>")
    if pos == -1:
        return None
    return pos + 4


def read_cintil_paragraphs(input_file, format_spec=None):
    paragraph_lines = []  # start reading new paragraph
    for linenum, line in enumerate(input_file, start=1):
        line = line.strip()
        if not line:  # skip empty lines
            continue
        if paragraph_lines:  # we are inside a paragraph
            close_p_pos = _get_close_p_pos(line)
            if close_p_pos is None:  # paragraph does not end on this line
                paragraph_lines.append(line)
                continue  # onto next line
            # else paragraph ends on close_p_pos
            paragraph_lines.append(line[:close_p_pos])
            paragraph = " ".join(paragraph_lines)
            yield parse_cintil_paragraph(
                paragraph, format_spec
            ) if format_spec else paragraph
            paragraph_lines = []  # start reading new paragraph from line[close_p_pos:]
            line = line[close_p_pos:].lstrip()
            if not line:
                continue  # onto next line
            # else fall through and start reading new paragraph from the
            # remainder of the current line
        while line:
            # handle the unusual situation of one or more <p> ... </p> in a single line
            if not line.startswith("<p>"):  # left space has been stripped above
                raise CintilExtraneousText(
                    f"line {linenum} has text outside <p> and </p> tags"
                )
            close_p_pos = _get_close_p_pos(line)
            if close_p_pos is None:
                paragraph_lines.append(line)
                # whole (remainder of) line is part of paragraph
                break  # nothing left to do with this line
            paragraph = line[:close_p_pos]
            yield parse_cintil_paragraph(
                paragraph, format_spec
            ) if format_spec else paragraph
            line = line[
                close_p_pos:
            ].strip()  # will continue the while loop until line is empty
    if paragraph_lines:
        raise CintilParagraphMissingTag("unclosed paragraph at the end of file")


def write_cintil_paragraphs(paragraphs, output_file, format_spec):
    for paragraph in paragraphs:
        print(format_cintil_paragraph(paragraph, format_spec), file=output_file)



"""This module implements the JSONL (one JSON object per Line) text annotation format.
 This is the preferred future-proof format.

 Each line contains a sentence, which is a JSON object containing the following fields:

 - tokens (mandatory/list of strings): tokens that compose the sentence
 - lemmas (optional/list of strings): lemma for each token
 - ulemmas (optional/list of strings): universal lemma for each token
 - pos_tags (optional/list of strings): part-of-speech tag for each token
 - upos_tags (optional/list of strings): universal part-of-speech tag for each token
 - ne_tags (optional/list of strings): named entity tag for each token
 - feats (optional/list of strings): morphological/inflection features for each token
 - ufeats (optional/list of strings): universal morphological/inflection features
                                      for each token
 - dep_heads (optional/list of ints): dependency relation head for each token
 - dep_rels (optional/list of strings): dependency relation tag for each token
 - udep_heads (optional/list of ints): universal dependency relation head for each
                                      token
 - udep_rels (optional/list of strings): universal dependency relation tag for each
                                      token
 - spaces (optional/string): one character for each token indicating spacing:
                              L=Left, R=Right, B=Both, N=None
 - par_id (optiona/string): the identifier of the paragraph to which this sentence
                            belongs
 - raw_sent (optional/string): the original plain-text sentence before tokenization
 - ctree (optional/string): constituency tree in parentheses format
 - sent_id (optional/string/int): sentence identifier
 - par_id (optional/string/int): paragraph identifier
"""

import json
import logging

from .lxparagraph import LxParagraph
from .lxsentence import LxSentence
from .lxtoken import LxToken

LOG = logging.getLogger(__name__)

JSONSENTENCE_LXTOKEN_MAPPING = [
    # (jsonsentence key, lxtoken attr)
    ("pos_tags", "pos"),
    ("upos_tags", "upos"),
    ("lemmas", "lemma"),
    ("ulemmas", "ulemma"),
    ("spaces", "space"),
    ("ne_tags", "ne"),
    ("ne_rb_tags", "ne_rb"),
    ("feats", "infl"),
    ("ufeats", "uinfl"),
    ("dep_heads", "parent"),
    ("dep_rels", "deprel"),
    ("udep_heads", "uparent"),
    ("udep_rels", "udeprel"),
]


def jsonsentence_to_lxsentence(jsonsentence):
    lxtokens = [LxToken(token) for token in jsonsentence["tokens"]]
    for jsonsentence_field, lxtoken_attr in JSONSENTENCE_LXTOKEN_MAPPING:
        if jsonsentence_field in jsonsentence:
            for value, lxtoken in zip(jsonsentence[jsonsentence_field], lxtokens):
                setattr(lxtoken, lxtoken_attr, value)
    raw = jsonsentence.get("raw_sent", None)
    sent_id = jsonsentence.get("sent_id", None)
    par_id = jsonsentence.get("par_id", None)
    ctree = jsonsentence.get("ctree", None)
    lxsentence = LxSentence(
        *lxtokens, raw=raw, ctree=ctree, sent_id=sent_id, par_id=par_id
    )
    return lxsentence


def lxsentence_to_jsonsentence(lxsentence):
    jsonsentence = dict(tokens=[token.form for token in lxsentence])
    for key, attr in JSONSENTENCE_LXTOKEN_MAPPING:
        values = [
            getattr(token, attr) if hasattr(token, attr) else None
            for token in lxsentence
        ]
        if any(value is not None for value in values):
            jsonsentence[key] = values
    for attr in ["ctree", "sent_id", "par_id"]:
        value = getattr(lxsentence, attr)
        if value is not None:
            jsonsentence[attr] = value
    return jsonsentence


def read_jsonl_sentences(input_file, ignore_errors=False):
    for linenum, line in enumerate(input_file, start=1):
        try:
            yield jsonsentence_to_lxsentence(json.loads(line))
        except json.JSONDecodeError:
            LOG.exception(f"invalid JSON object in line {linenum}")
            if ignore_errors:
                continue
            raise


def read_jsonl_paragraphs(input_file, ignore_errors=False):
    par, par_id = [], None
    for sentence in read_jsonl_sentences(input_file, ignore_errors):
        if hasattr(sentence, "par_id"):
            if par:
                if par_id is not None and par_id == sentence.par_id:
                    par.append(sentence)
                else:
                    yield LxParagraph(*par, par_id=par_id)
                    par, par_id = [sentence], sentence.par_id
            else:
                par.append(sentence)
        else:
            if par:
                yield LxParagraph(*par, par_id=par_id)
                par, par_id = [], None
            yield LxParagraph(sentence)
    if par:
        yield LxParagraph(*par, par_id=par_id)


def write_jsonl_sentence(lxsentence, output_file):
    jsonsentence = lxsentence_to_jsonsentence(lxsentence)
    print(json.dumps(jsonsentence), file=output_file)


def write_jsonl_paragraph(lxparagraph, output_file):
    for lxsentence in lxparagraph:
        write_jsonl_sentence(lxsentence, output_file)


def write_jsonl_paragraphs(lxparagraphs, output_file):
    for lxparagraph in lxparagraphs:
        write_jsonl_paragraph(lxparagraph, output_file)



import itertools
import unicodedata


def recase_as(s, r):
    if r.islower():
        return s.lower()
    if r.isupper():
        return s.upper()
    if r.istitle():
        return s.title()
    return s.capitalize()


# shamelessly copied from Python's documentation
# https://docs.python.org/3/library/itertools.html?highlight=itertools
def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


# shamelessly copied from Python's documentation
# https://docs.python.org/3/library/itertools.html?highlight=itertools
def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


def normalize_unicode(s):
    """
    Ensure that strings that look the same to a human reader compare equal.

    This one-liner function exists to ensure that the same normalization standard
     is applied everywhere.

    Don't use NFKC because it will replace "ª" with "a".

    See https://docs.python.org/3/library/unicodedata.html#unicodedata.normalize
    """
    return unicodedata.normalize("NFC", s)


def normalize_text(s):
    return normalize_unicode(s).replace("\xad", "-")


__all__ = ["recase_as", "pairwise", "grouper", "normalize_unicode"]



import re
from .cintilformat import (
    parse_cintil_sentence,
    format_cintil_sentence,
)
from .conllformat import format_conll_sentence
from .lxtoken import LxToken
from .penn import escape_token
from .utils import normalize_unicode

import lxcommon.conllx


class LxSentence(list):
    def __init__(
        self,
        *tokens,
        raw=None,
        default_cintil_format_spec=None,
        ctree=None,
        sent_id=None,
        par_id=None,
    ):
        if raw is not None:
            if not isinstance(raw, str):
                raise TypeError("raw must be a string")
        if not all(isinstance(token, LxToken) for token in tokens):
            raise TypeError("all tokens must be instances of LxToken")
        super().__init__(tokens)
        self.raw = normalize_unicode(raw) if isinstance(raw, str) else raw
        self.ctree = ctree
        self.default_cintil_format_spec = default_cintil_format_spec
        self.sent_id = sent_id
        self.par_id = par_id

    # TODO: maybe this property should be set when LxTokenizer is run on this sentence
    @property
    def tokenized(self):
        return len(self) > 0 or self.raw is None or self.raw.strip() == ""

    # TODO: maybe this property should be set when LxTagger is run on this sentence
    @property
    def tagged(self):
        return self.tokenized and all(token.pos is not None for token in self)

    # TODO: maybe this property should be set when LxSuite is run on this sentence
    @property
    def analysed(self):
        return (
            self.tokenized
            and self.tagged
            and (
                any(token.feats is not None for token in self)
                or any(token.lemma is not None for token in self)
            )
        )

    def upgrade_udeprels(self):
        "Upgrades Universal Dependency relations from V1 to V2"
        for token in self:
            token.upgrade_udeprel()
        return self

    def check_tagsets(self):
        sentence_errors = []
        for pos, token in enumerate(self, start=1):
            token_errors = token.check_tagsets()
            if token_errors:
                sentence_errors.append((pos, token_errors))
        return sentence_errors

    def get_ctree_pseudotags(self):
        pseudotags = []
        pattern = r"(?P<left>[^\)]+) (?P<token>[^\)]*)(?P<right>\)+)(?: |$)"
        ctree = getattr(self, "ctree", None)
        if not ctree:
            raise Exception("sentence has no ctree")
        matches = re.findall(pattern, ctree)
        if len(matches) != len(self):
            raise Exception(
                "ctree is inconsistent with sentence tokens; "
                f"ctree seems to have {len(matches)} tokens; expected {len(self)}"
            )
        for (left, form, right), token in zip(matches, self):
            if form != escape_token(token.form):
                raise Exception(
                    "ctree is inconsistent with sentence tokens; "
                    f"expected {token.form}, found {form}"
                )
            pseudotags.append(left.replace(" ", "") + "*" + right.replace(" ", ""))
        return pseudotags

    def to_plain_text(self):
        if self.raw:
            return self.raw
        buffer = []
        for token in self:
            if "L" in token.space:
                buffer.append(" ")
            if token.raw:
                buffer.append(token.raw)
        return "".join(buffer)

    def to_cintil(self, format_spec=None):
        if format_spec is None:
            if self.default_cintil_format_spec is None:
                raise ValueError(
                    "format_spec cannot be None "
                    "because self.default_cintil_format_spec is None"
                )
            format_spec = self.default_cintil_format_spec
        return format_cintil_sentence(self.to_primitive_types(), format_spec)

    def to_conll(self, universal_deps=False, universal=False, empty="-"):
        return format_conll_sentence(
            self.to_primitive_types(),
            universal_deps=universal_deps,
            universal=universal,
            empty=empty,
        )

    def to_conll_ctree(self):
        return "\n".join(
            [
                token.form + "\t" + pseudotag
                for token, pseudotag in zip(self, self.get_ctree_pseudotags())
            ]
        )

    @staticmethod
    def from_cintil(sentence, format_spec):
        return LxSentence.from_primitive_types(
            parse_cintil_sentence(sentence, format_spec),
            default_cintil_format_spec=format_spec,
        )

    @staticmethod
    def from_primitive_types(
        sentence, raw=None, default_cintil_format_spec=None, ctree=None
    ):
        "This method is useful when one needs to send/receive sentences via XML-RPC"
        if isinstance(sentence, list):
            return LxSentence(
                *[LxToken.from_primitive_types(token) for token in sentence],
                raw=raw,
                default_cintil_format_spec=default_cintil_format_spec,
                ctree=ctree,
            )
        if isinstance(sentence, str):
            return LxSentence(
                raw=sentence,
                default_cintil_format_spec=default_cintil_format_spec,
                ctree=ctree,
            )
        raise TypeError("sentence must be a list or a string")

    @staticmethod
    def from_conll(lines, universal=False):
        sentence = lxcommon.conllx.readsent(lines)
        if universal:
            return LxSentence(
                *[
                    LxToken(
                        t.form,
                        ulemma=t.lemma,
                        upos=t.pos,
                        ufeats=t.feats,
                        uparent=t.head,
                        udeprel=t.deprel,
                    )
                    for t in sentence
                ]
            )
        return LxSentence(
            *[
                LxToken(
                    t.form,
                    lemma=t.lemma,
                    pos=t.pos,
                    feats=t.feats,
                    parent=t.head,
                    deprel=t.deprel,
                )
                for t in sentence
            ]
        )

    def to_primitive_types(self, **kwargs):
        "This method is useful when one needs to send/receive sentences via XML-RPC"
        if self.tokenized:
            return [token.to_primitive_types(**kwargs) for token in self]
        return self.raw

    def __repr__(self):
        if not self.tokenized:
            if self.raw:
                return f"LxSentence(raw={self.raw!r})"
            else:
                return "LxSentence()"
        tokens = ", ".join([repr(token) for token in self])
        if self.raw:
            return f"LxSentence({tokens}, raw={self.raw!r})"
        return f"LxSentence({tokens})"

    def copy(self):
        return LxSentence(
            *[token.copy() for token in self],
            raw=self.raw,
            default_cintil_format_spec=self.default_cintil_format_spec,
        )



"""Module for reading and writing CONLL-X format

 Luís Gomes <luis.gomes@di.fc.ul.pt>, <luismsgomes@gmail.com>

 Copyright NLX-Group, Universidade de Lisboa, 2015-2022


The following text is adapted from http://ilk.uvt.nl/conll/:

Files contain sentences separated by a blank line.
A sentence consists of one or tokens, each one starting on a new line.
A token consists of ten fields described in the table below. Fields are
 separated by a single tab character. Space/blank characters are not allowed in
 within fields.
All files will contains these ten fields, although only the ID, FORM, CPOSTAG,
 POSTAG, HEAD and DEPREL columns are guaranteed to contain non-dummy (i.e.
 non-underscore) values.

Fields:
1   ID      Token counter, starting at 1 for each new sentence.
2   FORM    Word form or punctuation symbol.
3   LEMMA   Lemma or stem (depending on particular data set) of word form, or an
            underscore if not available.
4   CPOSTAG Coarse-grained part-of-speech tag, where tagset depends on the
            language.
5   POSTAG  Fine-grained part-of-speech tag, where the tagset depends on the
            language, or identical to the coarse-grained part-of-speech tag if
            not available.
6   FEATS   Unordered set of syntactic and/or morphological features (depending
            on the particular language), separated by a vertical bar (|), or an
            underscore if not available.
7   HEAD    Head of the current token, which is either a value of ID or zero
            ('0'). Note that depending on the original treebank annotation,
            there may be multiple tokens with an ID of zero.
8   DEPREL  Dependency relation to the HEAD. The set of dependency relations
            depends on the particular language. Note that depending on the
            original treebank annotation, the dependency relation may be
            meaningfull or simply 'ROOT'.
9   PHEAD   Projective head of current token, which is either a value of ID or
            zero ('0'), or an underscore if not available. Note that depending
            on the original treebank annotation, there may be multiple tokens an
            with ID of zero. The dependency structure resulting from the PHEAD
            column is guaranteed to be projective (but is not available for all
            languages), whereas the structures resulting from the HEAD column
            will be non-projective for some sentences of some languages (but is
            always available).
10  PDEPREL Dependency relation to the PHEAD, or an underscore if not available.
            The set of dependency relations depends on the particular language.
            Note that depending on the original treebank annotation, the
            dependency relation may be meaningfull or simply 'ROOT'.
"""

import sys


EMPTY_MARKERS = (
    "_",  # this is the standard marker for no value
    "-",
    "",
)


class ConllxToken:
    __slots__ = [  # field
        "tokid",  # 1
        "form",  # 2
        "lemma",  # 3
        "cpos",  # 4
        "pos",  # 5
        "feats",  # 6
        "head",  # 7
        "deprel",  # 8
        "phead",  # 9
        "pdeprel",  # 10
    ]

    def __init__(self, *args, **kwargs):
        """Creates a token from str, tuple or dict"""
        if args:
            if len(args) == 1:
                if isinstance(args[0], str):
                    self._init_from_str(args[0])
                elif isinstance(args[0], tuple):
                    self._init_from_seq(args[0])
                elif isinstance(args[0], dict):
                    self._init_from_dict(args[0])
                else:
                    raise ValueError("Unexpected argument type: " + type(args[0]))
            else:
                assert len(args) == len(ConllxToken.__slots__)
                self._init_from_seq(args)
        elif kwargs:
            self._init_from_dict(kwargs)
        else:
            raise ValueError("Missing argument(s)")

    def __str__(self):
        return self.asstr()

    def __repr__(self):
        return f"ConllxToken{self.astuple()!r}"

    def asstr(self):
        return "\t".join(
            [EMPTY_MARKERS[0] if v is None else str(v) for v in self.astuple()]
        )

    def astuple(self):
        return tuple(getattr(self, attr) for attr in ConllxToken.__slots__)

    def asdict(self):
        return {attr: getattr(self, attr) for attr in ConllxToken.__slots__}

    def _init_from_str(self, arg):
        fields = arg.split("\t")
        assert len(fields) == 10
        try:
            fields = tuple(
                None
                if field in EMPTY_MARKERS and fieldnum not in {2, 3}  # form, lemma
                else int(field)
                if fieldnum in {1, 7, 9}  # tokid, head, phead
                else field
                for fieldnum, field in enumerate(fields, start=1)
            )
        except Exception as ex:
            print("Invalid line:", arg, file=sys.stderr)
            print("Exception:", ex, file=sys.stderr)
            raise ex
        self._init_from_seq(fields)

    def _init_from_seq(self, arg):
        for attr, v in zip(ConllxToken.__slots__, arg):
            setattr(self, attr, v)

    def _init_from_dict(self, arg):
        for attr in ConllxToken.__slots__:
            setattr(self, attr, arg.get(attr, None))


def readsent(f):
    sent = []
    for line in f:
        line = line.rstrip("\n")
        if line.startswith("#"):
            continue
        if line:
            sent.append(ConllxToken(line))
        elif len(sent) > 0:
            break
    return sent


def writesent(f, sent, flush=True):
    assert isinstance(sent, list) and len(sent) > 0
    for token in sent:
        assert isinstance(token, ConllxToken)
        print(str(token), file=f)
    print(flush=flush, file=f)


def read(f):
    while True:
        sent = readsent(f)
        if sent:
            yield sent
        else:
            return


def write(f, sents, flush=True):
    for sent in sents:
        writesent(f, sent, flush=flush)


def fix_roots(sent):
    rootid = None
    assert all(isinstance(token, ConllxToken) for token in sent)
    for tok in sent:
        if tok.head == 0 and tok.deprel == "ROOT":
            rootid = tok.tokid
            break
    assert rootid is not None
    for tok in sent:
        if tok.tokid != rootid and tok.head == 0:
            tok.head = rootid
            if tok.phead is not None:
                tok.phead = tok.head
            if tok.deprel == "ROOT":
                tok.deprel = "DEP"
                if tok.pdeprel is not None:
                    tok.pdeprel = tok.deprel
    return sent


__all__ = ["ConllxToken", "readsent", "writesent", "read", "write", "fix_roots"]



from .cintilformat import (
    parse_cintil_paragraph,
    read_cintil_paragraphs,
    format_cintil_paragraph,
)
from .lxsentence import LxSentence


class LxParagraph(list):
    def __init__(self, *sentences, default_cintil_format_spec=None, par_id=None):
        if not all(isinstance(sentence, LxSentence) for sentence in sentences):
            raise TypeError("all sentences must be instances of LxSentence")
        super().__init__(sentences)
        self.default_cintil_format_spec = default_cintil_format_spec
        self.par_id = par_id

    def __repr__(self):
        sentences = ", ".join([repr(s) for s in self])
        return f"LxParagraph({sentences})"

    @staticmethod
    def from_primitive_types(paragraph, default_cintil_format_spec=None):
        "This method is useful when one needs to send/receive paragraphs via XML-RPC"
        if not isinstance(paragraph, list):
            raise TypeError("argument must be a list")
        return LxParagraph(
            *[
                LxSentence.from_primitive_types(
                    sentence, default_cintil_format_spec=default_cintil_format_spec
                )
                for sentence in paragraph
            ],
            default_cintil_format_spec=default_cintil_format_spec,
        )

    def to_primitive_types(self):
        "This method is useful when one needs to send/receive paragraphs via XML-RPC"
        return [sentence.to_primitive_types() for sentence in self]

    def to_plain_text(self, sentence_sep="\n"):
        if not isinstance(sentence_sep, str):
            raise TypeError("sentence_sep must be a string")
        return sentence_sep.join([sentence.to_plain_text() for sentence in self])

    @staticmethod
    def from_cintil(paragraph, format_spec):
        return LxParagraph.from_primitive_types(
            parse_cintil_paragraph(paragraph, format_spec),
            default_cintil_format_spec=format_spec,
        )

    def to_cintil(self, format_spec=None):
        if format_spec is None:
            if self.default_cintil_format_spec is None:
                raise ValueError(
                    "format_spec cannot be None "
                    "because self.default_cintil_format_spec is None"
                )
            format_spec = self.default_cintil_format_spec
        return format_cintil_paragraph(self.to_primitive_types(), format_spec)

    @staticmethod
    def read_cintil_paragraphs(input_file, format_spec):
        for paragraph in read_cintil_paragraphs(input_file, format_spec):
            yield LxParagraph.from_primitive_types(
                paragraph, default_cintil_format_spec=format_spec
            )

    def write_cintil(self, output_file, format_spec):
        print(self.to_cintil(format_spec), file=output_file)

    def to_ctrees(self):
        for sentence in self:
            yield getattr(sentence, "ctree", "")

    def write_ctrees(self, output_file):
        for ctree in self.to_ctrees():
            print(ctree, file=output_file)
        print(file=output_file)

    def to_conll_ctrees(self):
        for sentence in self:
            yield sentence.to_conll_ctree()

    def write_conll_ctrees(self, output_file):
        for conll_ctree in self:
            print(conll_ctree, end="\n\n", file=output_file)

    def copy(self):
        return LxParagraph(*[sentence.copy() for sentence in self])


__all__ = ["LxParagraph"]



CONLL_HEADER = "\t".join(
    [
        "id",  # 1
        "form",  # 2
        "lemma",  # 3
        "cpos",  # 4
        "pos",  # 5
        "feat",  # 6
        "head",  # 7
        "deprel",  # 8
        "phead",  # 9
        "pdeprel",  # 10
    ]
)


def format_conll_sentence(sentence, universal_deps=False, universal=False, empty="-"):
    if not isinstance(sentence, (list, tuple)):
        raise TypeError("sentence must be an instance of list or tuple")
    if not all(isinstance(token, dict) for token in sentence):
        raise TypeError("each token must be an instance of dict")
    lines = ["#" + CONLL_HEADER]
    for tokid, token in enumerate(sentence, start=1):
        tokid = str(tokid)
        form = token.get("form")
        lemma = token.get("ulemma" if universal else "lemma", empty)
        cpos = token.get("upos" if universal else "pos", empty)
        pos = cpos
        feats = token.get(
            "ufeats" if universal else "feats",
            token.get("uinfl" if universal else "infl", empty),
        )
        head = str(
            token.get("uparent" if universal or universal_deps else "parent", empty)
        )
        deprel = token.get(
            "udeprel" if universal or universal_deps else "deprel", empty
        )
        phead = head
        pdeprel = deprel

        lines.append(
            "\t".join(
                [
                    tokid,  # 1
                    form,  # 2
                    lemma,  # 3
                    cpos,  # 4
                    pos,  # 5
                    feats,  # 6
                    head,  # 7
                    deprel,  # 8
                    phead,  # 9
                    pdeprel,  # 10
                ]
            )
        )
    lines.append("")  # empty line marks end of sentence
    return "\n".join(lines) + "\n"  # end with newline



"""
Luís Gomes <luis.gomes@di.fc.ul.pt>, <luismsgomes@gmail.com>

Copyright NLX-Group, Universidade de Lisboa, 2015-2022
"""

import logging

from .cintilformat import (
    parse_cintil_token,
    format_cintil_token,
)

from .utils import normalize_unicode, recase_as
from .tagsets import (
    LX_POS_TAGS,
    U_POS_TAGS,
    LX_DEPREL_TAGS,
    U_DEPREL_TAGS,
    U_DEPREL_TAGS_V1_TO_V2,
    LX_FEATS,
    U_FEATS,
    NE_TAGS,
)


LOG = logging.getLogger(__name__)


class LxToken:
    _attrs = [
        "form",
        "raw",
        "lemma",
        "pos",
        "_upos",
        "feats",
        "ufeats",
        "ne",
        "ne_rb",
        "deprel",
        "parent",
        "_udeprel",
        "uparent",
        "space",
    ]

    def __init__(
        self,
        form,
        space="LR",
        raw=None,
        pos=None,
        upos=None,
        lemma=None,
        ulemma=None,
        feats=None,
        ufeats=None,
        infl=None,
        uinfl=None,
        ne=None,
        ne_rb=None,
        deprel=None,
        parent=None,
        udeprel=None,
        uparent=None,
        **kwargs,
    ):
        self.form = normalize_unicode(form)
        self.space = ""
        self.add_space(space)
        self.raw = normalize_unicode(raw) if isinstance(raw, str) else raw
        self.pos = pos
        self.upos = upos
        self.lemma = lemma
        self.ulemma = ulemma
        if feats is not None:
            if infl is not None and infl != feats:
                raise ValueError(
                    "Attributes feats and infl cannot be given simultaneously "
                    "and with different values."
                )
        if ufeats is not None:
            if uinfl is not None and uinfl != ufeats:
                raise ValueError(
                    "Attributes ufeats and uinfl cannot be given simultaneously "
                    "and with different values."
                )
        self.feats = feats or infl
        self.ufeats = ufeats or uinfl
        self.ne = ne
        self.ne_rb = ne_rb
        self.deprel = deprel
        self.parent = parent
        self.udeprel = udeprel
        self.uparent = uparent
        for name, value in kwargs.items():
            setattr(self, name, value)

    @property
    def deprel(self):
        return self._deprel

    @deprel.setter
    def deprel(self, deprel):
        self._deprel = deprel.upper() if deprel else None

    @property
    def pos(self):
        return self._pos

    @pos.setter
    def pos(self, pos):
        self._pos = pos.upper() if pos else None

    @property
    def udeprel(self):
        return self._udeprel

    @udeprel.setter
    def udeprel(self, udeprel):
        self._udeprel = udeprel.lower() if udeprel else None

    @property
    def upos(self):
        return self._upos

    @upos.setter
    def upos(self, upos):
        self._upos = upos.upper() if upos else None

    @property
    def infl(self):
        return self.feats

    @infl.setter
    def infl(self, feats):
        self.feats = feats

    def __eq__(self, other):
        return isinstance(other, LxToken) and self.__dict__ == other.__dict__

    def add_space(self, where):
        for c in where.upper():
            if c not in ["L", "R"]:
                raise ValueError(f"Invalid character {c!r} in where")
            if c not in self.space:
                if c == "L":
                    self.space = c + self.space
                else:
                    self.space = self.space + c

    def remove_space(self, where):
        for c in where.upper():
            if c not in ["L", "R"]:
                raise ValueError(f"Invalid character {c!r} in where")
            if c in self.space:
                self.space = self.space.replace(c, "")

    @staticmethod
    def from_primitive_types(token):
        if not isinstance(token, dict):
            raise TypeError("token must be a dict")
        form = token["form"]
        attrs = {a: v if v != "_" else None for a, v in token.items() if a != "form"}
        return LxToken(form, **attrs)

    def to_primitive_types(self, empty_attrs=False, extra_attrs=True, infl=False):
        attr_name_mapping = {
            "_pos": "pos",
            "_deprel": "deprel",
            "_upos": "upos",
            "_udeprel": "udeprel",
        }
        if infl:
            attr_name_mapping["feats"] = "infl"
        if extra_attrs:
            return {
                attr_name_mapping.get(attr, attr): value
                for attr, value in self.__dict__.items()
                if empty_attrs or value is not None
            }
        return {
            attr_name_mapping.get(attr, attr): getattr(self, attr, None)
            for attr in LxToken._attrs
            if empty_attrs or getattr(self, attr, None) is not None
        }

    def to_cintil(self, format_spec):
        return format_cintil_token(self.to_primitive_types(), format_spec)

    @staticmethod
    def from_cintil(token, format_spec):
        return LxToken.from_primitive_types(parse_cintil_token(token, format_spec))

    def copy(self):
        return LxToken(
            self.form,
            **{attr: value for attr, value in self.__dict__.items() if attr != "form"},
        )

    def strip_pos(self):
        self.pos = None

    def strip_upos(self):
        self.upos = None

    def strip_morph(self):
        self.lemma = None
        self.feats = None

    def strip_umorph(self):
        self.ulemma = None
        self.ufeats = None

    def strip_ne(self):
        self.ne = None

    def strip_ne_rb(self):
        self.ne_rb = None

    def strip_dep(self):
        self.deprel = None
        self.parent = None

    def strip_udep(self):
        self.udeprel = None
        self.uparent = None

    def istag(self):
        # this may give false positives but it's better to be on the safe side
        return self.form.startswith("<") and self.form.endswith(">")

    def upgrade_udeprel(self):
        if self.udeprel in U_DEPREL_TAGS_V1_TO_V2:
            self.udeprel = U_DEPREL_TAGS_V1_TO_V2[self.udeprel]

    def check_tagsets(self):
        errors = {}
        if self.pos and self.pos not in LX_POS_TAGS:
            errors["pos"] = self.pos
        if self.upos and self.upos not in U_POS_TAGS:
            errors["upos"] = self.upos

        if self.deprel and self.deprel not in LX_DEPREL_TAGS:
            errors["deprel"] = self.deprel
        if self.udeprel and self.udeprel not in U_DEPREL_TAGS:
            errors["udeprel"] = self.udeprel

        if self.feats and self.feats not in LX_FEATS:
            errors["feats"] = self.feats
        if self.ufeats and self.ufeats not in U_FEATS:
            errors["ufeats"] = self.ufeats

        if self.ne and self.ne not in NE_TAGS:
            errors["ne"] = self.ne
        return errors

    def __repr__(self):
        attr_name_mapping = {
            "_pos": "pos",
            "_deprel": "deprel",
            "_upos": "upos",
            "_udeprel": "udeprel",
        }
        attrs = ", ".join(
            [
                f"{attr_name_mapping.get(attr, attr)}={value!r}"
                for attr, value in sorted(self.__dict__.items())
                if attr != "form" and value is not None
            ]
        )
        return f"LxToken({self.form!r}, {attrs})"

    @property
    def recased_form(self):
        return recase_as(self.form, self.raw) if self.raw else self.form


__all__ = ["LxToken"]



CONTRACTIONS = {
    # expanded => contracted form (lowercased)
    # list by João Silva:
    "a_ a": "à",
    "a_ as": "às",
    "a_ o": "ao",
    "a_ os": "aos",
    "a_ aquela": "àquela",
    "a_ aquelas": "àquelas",
    "a_ aquele": "àquele",
    "a_ aqueles": "àqueles",
    "a_ onde": "aonde",
    "com_ mim": "comigo",
    "com_ ti": "contigo",
    "com_ nós": "connosco",
    "com_ vós": "convosco",
    "com_ si": "consigo",
    "de_ a": "da",
    "de_ as": "das",
    "de_ o": "do",
    "de_ os": "dos",
    "de_ uma": "duma",
    "de_ umas": "dumas",
    "de_ um": "dum",
    "de_ uns": "duns",
    "de_ ela": "dela",
    "de_ elas": "delas",
    "de_ ele": "dele",
    "de_ eles": "deles",
    "de_ entre": "dentre",
    "de_ esta": "desta",
    "de_ estas": "destas",
    "de_ este": "deste",
    "de_ estes": "destes",
    "de_ essa": "dessa",
    "de_ essas": "dessas",
    "de_ esse": "desse",
    "de_ esses": "desses",
    "de_ isto": "disto",
    "de_ isso": "disso",
    "de_ aquela": "daquela",
    "de_ aquelas": "daquelas",
    "de_ aquele": "daquele",
    "de_ aqueles": "daqueles",
    "de_ aquilo": "daquilo",
    "de_ aqueloutros": "daqueloutros",
    "de_ aí": "daí",
    "de_ ali": "dali",
    "de_ aqui": "daqui",
    "de_ onde": "donde",
    "de_ antes": "dantes",
    "em_ a": "na",
    "em_ as": "nas",
    "em_ o": "no",
    "em_ os": "nos",
    "em_ uma": "numa",
    "em_ umas": "numas",
    "em_ um": "num",
    "em_ uns": "nuns",
    "em_ ela": "nela",
    "em_ elas": "nelas",
    "em_ ele": "nele",
    "em_ eles": "neles",
    "em_ esta": "nesta",
    "em_ estas": "nestas",
    "em_ este": "neste",
    "em_ estes": "nestes",
    "em_ essa": "nessa",
    "em_ essas": "nessas",
    "em_ esse": "nesse",
    "em_ esses": "nesses",
    "em_ outra": "noutra",
    "em_ outras": "noutras",
    "em_ outro": "noutro",
    "em_ outros": "noutros",
    "em_ isto": "nisto",
    "em_ isso": "nisso",
    "em_ aquela": "naquela",
    "em_ aquelas": "naquelas",
    "em_ aquele": "naquele",
    "em_ aqueles": "naqueles",
    "em_ aquilo": "naquilo",
    "esta_ outra": "estoutra",
    "por_ a": "pela",
    "por_ as": "pelas",
    "por_ o": "pelo",
    "por_ os": "pelos",
    # handle some cases using old orthography (from the early 20th century):
    "de_ èles": "dèles",  # ?
    "de_ êla": "dêla",
    "de_ êlas": "dêlas",
    # "de_ êle": "dê",  # repeated key further ahead
    # "de_ êles": "dê",  # repeated key further ahead
    "de_ êsta": "dêsta",
    "de_ êstas": "dêstas",
    "de_ êste": "dêste",
    "de_ êstes": "dêstes",
    "de_ êssa": "dêssa",
    "de_ êssas": "dêssas",
    "de_ êsse": "dêsse",
    "de_ êsses": "dêsses",
    "de_ êla": "dêla",
    "de_ êlas": "dêlas",
    "de_ êle": "dêle",
    "de_ êles": "dêles",
    "de_ aquêles": "daquêles",
    "em_ êla": "nêla",
    "em_ êlas": "nêlas",
    "em_ êle": "nêle",
    "em_ êles": "nêles",
    # handle contraction of clitic pronouns:
    "me_ a": "ma",
    "me_ as": "mas",
    "me_ o": "mo",
    "me_ os": "mos",
    "me_ -a": "ma",
    "me_ -as": "mas",
    "me_ -o": "mo",
    "me_ -os": "mos",
    "te_ a": "ta",
    "te_ as": "tas",
    "te_ o": "to",
    "te_ os": "tos",
    "te_ -a": "ta",
    "te_ -as": "tas",
    "te_ -o": "to",
    "te_ -os": "tos",
    "lhe_ a": "lha",
    "lhe_ as": "lhas",
    "lhe_ o": "lho",
    "lhe_ os": "lhos",
    "lhe_ -a": "lha",
    "lhe_ -as": "lhas",
    "lhe_ -o": "lho",
    "lhe_ -os": "lhos",
    # added by Luis Gomes:
    "-me_ a": "-ma",
    "-me_ as": "-mas",
    "-me_ o": "-mo",
    "-me_ os": "-mos",
    "-me_ -a": "-ma",
    "-me_ -as": "-mas",
    "-me_ -o": "-mo",
    "-me_ -os": "-mos",
    "-te_ a": "-ta",
    "-te_ as": "-tas",
    "-te_ o": "-to",
    "-te_ os": "-tos",
    "-te_ -a": "-ta",
    "-te_ -as": "-tas",
    "-te_ -o": "-to",
    "-te_ -os": "-tos",
    "-lhe_ a": "-lha",
    "-lhe_ as": "-lhas",
    "-lhe_ o": "-lho",
    "-lhe_ os": "-lhos",
    "-lhe_ -a": "-lha",
    "-lhe_ -as": "-lhas",
    "-lhe_ -o": "-lho",
    "-lhe_ -os": "-lhos",
    "a_ aquilo": "àquilo",
    "de_ outra": "doutra",
    "de_ outras": "doutras",
    "de_ outro": "doutro",
    "de_ outros": "doutros",
    "em_ alguns": "nalguns",
    "em_ algum": "nalgum",
    "em_ algumas": "nalgumas",
    "em_ alguma": "nalguma",
}



from setuptools import setup


setup(
    name="lxcommon",
    author="Luís Gomes",
    author_email="luis.gomes@di.fc.ul.pt",
    version="2.13.0",
    description="LX-Common is a package for code bits shared across several LX tools.",
    package_dir={"": "src"},
    packages=["lxcommon"],
)
